//
// THIS FILE IS COPYRIGHT 1994, 1995 BY TELENOR R&D, NORWAY
//
//                        ALL RIGHTS RESERVED.
//
// For rights granted to the public see the COPYRIGHT file
//
//
// Disk.cc	Handles all files on disk. Writes out words to index
//	     	files, merges index files and searches index files.
//
// Author:  Ken Ronny Schouten
//
// Changes: January 1995 Brd Hfjeld
//				Changed disk block layout to match physical
//				    disk blocks
//			       	Added truncated search
//				Added date as search criteria
//				Major rewrite of disk handling
//				    functions for modularity
//				Merge totally rewritten for performance
//	    February 1995 Brd Hfjeld
//				Added support for date index
//
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "Disk.h"
#include "Index.h"
#include "div.h"
#include "lang.h"

#define True (1)
#define False (0) 

int Disk::MakeIndexOnDisk(Index& index)
{
    if (_mode == DREAD) {
	_error = DMODE;
	return -1;
    }

    // Store the url file first so that we have the offsets that
    // we need later. 
    StoreUrl(index._url);

    index._entry.travers(this);
    
    // Let the rest be stored on disk also.
    FlushBuffer();
    if (StoreHeader() == -1)
	return -1;

    // Generate the date index
    if ( MakeDateIndex(index) < 0 ) return -1;
    
    return 0;
}

int Disk::Merge(Disk& d1, Disk& d2)
{
    long offset = 0;
    int ret;
    unsigned char *w1, *w2;
    unsigned long o1, o2;	// File offsets
    unsigned long p1, p2;	// Number of pointers
    long t1, t2;		// Dates
    
    // Concatenate the url files
    while (!feof(d1._fp_url)) {
	ret = fread(d1._buffer, 1, _header.bsize, d1._fp_url);
	fwrite(d1._buffer, ret, 1, _fp_url);
    }
    offset = ftell(_fp_url);

    while (!feof(d2._fp_url)) {
	ret = fread(d2._buffer, 1, _header.bsize, d2._fp_url);
	fwrite(d2._buffer, ret, 1, _fp_url);
    }
    

    // Then walk through the word files and merge them
    // New offsets for d2 url's are calculated by adding
    // length of d1 url file

    d1.GetIndexBlock(0);
    d2.GetIndexBlock(0);

    if ( d1.error()!=DOK || d2.error()!=DOK ) return -1; 
    
    d1.GetNextIndexEntry(&w1, &o1);
    d2.GetNextIndexEntry(&w2, &o2);
    
    while ((d1.error() != DEOF) || (d2.error() != DEOF)) {
	if ((d1.error() == DEOF) && (d2.error() != DEOF)) {
	    p2 = d2.GetNumberOfPointers(o2);
	    CopyPointers(0, NULL, p2, d2._fp_ptr, offset);
	    SetNextIndexEntry(w2, pointer_fp);
	    pointer_fp += (p2 + 1) * sizeof(long);
	    d2.pointer_fp += p2 * sizeof(long);
	    d2.GetNextIndexEntry(&w2, &o2);
	} else if ((d1._error != DEOF) && (d2._error == DEOF)) {
	    p1 = d1.GetNumberOfPointers(o1);
	    CopyPointers(p1, d1._fp_ptr, 0, NULL, 0);
	    SetNextIndexEntry(w1, pointer_fp);
	    pointer_fp += (p1 + 1) * sizeof(long);
	    d1.pointer_fp += p1 * sizeof(long);
	    d1.GetNextIndexEntry(&w1, &o1);
	} else 	if ( strcmp8(w1, w2) < 0) {
	    p1 = d1.GetNumberOfPointers(o1);
	    CopyPointers(p1, d1._fp_ptr, 0, NULL, 0);
	    SetNextIndexEntry(w1, pointer_fp);
	    pointer_fp += (p1 + 1) * sizeof(long);
	    d1.pointer_fp += p1 * sizeof(long);
	    d1.GetNextIndexEntry(&w1, &o1);
	} else if ( strcmp8(w1, w2) == 0 ) {
	    p1 = d1.GetNumberOfPointers(o1);
	    p2 = d2.GetNumberOfPointers(o2);
	    CopyPointers(p1, d1._fp_ptr, p2, d2._fp_ptr, offset);
	    SetNextIndexEntry(w1, pointer_fp);
	    pointer_fp += (p1 + p2 + 1) * sizeof(long);
	    d1.pointer_fp += p1 * sizeof(long);
	    d2.pointer_fp += p2 * sizeof(long);
	    d1.GetNextIndexEntry(&w1, &o1);
	    d2.GetNextIndexEntry(&w2, &o2);
	} else {
	    p2 = d2.GetNumberOfPointers(o2);
	    CopyPointers(0, NULL, p2, d2._fp_ptr, offset);
	    SetNextIndexEntry(w2, pointer_fp);
	    pointer_fp += (p2 + 1) * sizeof(long);
	    d2.pointer_fp += p2 * sizeof(long);
	    d2.GetNextIndexEntry(&w2, &o2);
	}
	
	// Check if a real error occurred anywhere
	if ( (d1.error() != DOK && d1.error() != DEOF) ||
	     (d2.error() != DOK && d2.error() != DEOF) ||
	     (error() != DOK ) )
	    return -1;
    }

    // Write last (partial) buffer to file
    if ( FlushBuffer() < 0 ) return -1;

    // Update header
    if (StoreHeader() < 0) return -1;

    // Check if both have date indexes, if not remove
    // dix file and return
    // For compatibility with no-dix indices
    if ( d1._fp_dix == NULL || d2._fp_dix == NULL ) {
	unsigned char filename[200];
	
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".dix", 4);

	unlink((char *)filename);
	
	return 0;
    }

    // Then merge the date index files
    d1._error = d2._error = _error = DOK;
    d1.GetNextDixEntry(&t1, &o1);
    d2.GetNextDixEntry(&t2, &o2);
    while ((d1.error() != DEOF) || (d2.error() != DEOF)) {
	if ( d2.error() == DEOF || (d1.error() == DOK && t1 <= t2) ) {
	    SetNextDixEntry(t1, o1);
    	    d1.GetNextDixEntry(&t1, &o1);
	} else {
	    SetNextDixEntry(t2, o2+offset);
    	    d2.GetNextDixEntry(&t2, &o2);
	}

	// Check if a real error occurred anywhere
	if ( (d1.error() != DOK && d1.error() != DEOF) ||
	     (d2.error() != DOK && d2.error() != DEOF) ||
	     (error() != DOK ) )
	    return -1;
    }
    
    return 0;
}

//
// Decide if a word might be in a block
//
int Disk::WordInBlock(unsigned char *word, int truncated)
{
    unsigned char *word1, *word2;

    word1 = &_buffer[_ind];
    word2 = &_buffer[_last];

    // Use strncmp if truncated search, else strcmp
    if ( truncated ) {
	int length = strlen((char *)word);
	if (strncmp8(word1, word, length) >= 0)
	    return -1;
    	else if (strncmp8(word2, word, length) < 0)
	    return 1;
    	else return 0;	
    } else {
	if (strcmp8(word1, word) > 0)
	    return -1;
    	else if (strcmp8(word2, word) < 0)
	    return 1;
    	else return 0;
    }

    return 0;
}


//
// Search according to the given criteria, return list of URLs
// This is the main search algorithm. Do search based on dates
// if word is "*", otherwise search based on the given word.
//
URLPTRSplaySet *Disk::Search(unsigned char *word,
			     struct search_params *sp)
{
    if ( strcmp8(word, (unsigned char *)"*") == 0 ) 
	return SearchForDate(sp);
    else
	return SearchForWord(word, sp);
}

//
// Search for the given word(s), return list of URLs
// This is the main search algorithm
//
URLPTRSplaySet *Disk::SearchForWord(unsigned char *searchword,
			      struct search_params *sp)
{
    int left;
    int right;
    int ret;
    int ind;
    int loop = 1;
    int truncated = 0;
    unsigned char *asterisk;
    int wordlength;
    unsigned char *word = (unsigned char *)strdup((char *)searchword);

    URLPTRSplaySet* urllist = new URLPTRSplaySet;

    if (_mode == DWRITE) {
	_error = DMODE;
	return NULL;
    }

    // Check for * in word and set truncated flag
    if ( asterisk = (unsigned char *)strchr((char *)word, '*') ) {
	if ( *(asterisk+1) != '\0' ) {
	    cout.form(BadTruncation, word);
	    exit(1);
	}	
	if ( asterisk == word ) {
	    cout.form(InternalError, "* passed to Disk::SearchForWord()");
	    exit(1);
	}	    
	*asterisk = '\0';
	truncated = True;
    }

    wordlength = strlen((char *)word);

    // First word in search word is index into hash array
    // Set low search limit from hash table
    if (_header.letter[word[0]] == -1) {
	return urllist;
    }
    else
	left = _header.letter[word[0]];

    // Set high search limit, default to last block
    right = _header.blocks - 1;
    for (int i = word[0]+1; i < 256; i++)
	if (_header.letter[i] != -1) {
	    right = _header.letter[i];
	    break;
	}

    // Binary search over the diskblocks
    while (loop) {

	ind = (int) (left + right) / 2;

	if (left == right)
	    loop = 0;

	// Load the block
	GetIndexBlock(ind);

	// Check if word is in loaded block
	ret = WordInBlock(word, truncated);

	if (ret == 0) 
	    break;
	else if (ret == -1) {
	    right = ind-1;
	    if ( right < left ) right = left;
	} else {
	    left = ind+1;
	    if ( left > right ) left = right;
	}
    }

    // Search failed if exact, return empty urllist
    // If truncated search we might be one block ahead true data,
    // so we have to scan.
    if (ret != 0 && !truncated )
	return urllist;

    // OK, we have a block.. now find first matching word in it if any
    int result;			// Result of last comparison
    unsigned char *eword;	// Current entry word
    unsigned long offset;	// Current ptr file offset

    // Linear search for the word within the block 
    while (_ind <= _last) {
	if ( GetNextIndexEntry(&eword, &offset) < 0 ) return NULL;

	if ( truncated ) {
	    result = strncmp8(word, eword, wordlength);
	} else {
	    result = strcmp8(word, eword);
	}
	
        // If we are past the wanted word
	if (result < 0) {
	    return urllist;
	}

	// If word matches
	if (result == 0) break;
    }

    // If last result was not a match search failed
    if ( result != 0 ) return urllist;
    
    // When we get here (eword, offset) is first matching entry

    // For untruncated searches just grab the entry and return
    if ( !truncated ) {
	if ( GetUrls(urllist, offset, sp) < 0 ) return NULL;

	// Check if there are too many entries..
	if ( urllist->length() > sp->max_internal ) {
	    cout.form(InternalExceeded, sp->max_internal);
	    exit(1);
	}
	
	return urllist;
    }
    
    // Truncated searches have to scan until no more matches
    while ( result == 0 ) {	
	if ( GetUrls(urllist, offset, sp) < 0 ) return NULL;

	// Check if there are too many entries..
	if ( urllist->length() > sp->max_internal ) {
	    cout.form(InternalExceeded, sp->max_internal);
	    exit(1);
	}

	// Next index entry
	if ( GetNextIndexEntry(&eword, &offset) < 0 ) {
	    if ( _error == DEOF ) return urllist;
	    else return NULL;
	}
	    
	result = strncmp8(word, eword, wordlength);
    }

    return urllist;
}

//
// Search for documents between the given dates, return list of URLs
//
URLPTRSplaySet *Disk::SearchForDate(struct search_params *sp)
{
    unsigned long low, split, high;
    struct stat stats;
    long date;
    unsigned long offset;
    Url *url;
    
    URLPTRSplaySet* urllist = new URLPTRSplaySet;

    // First decide number of entries in dix
    if ( _fp_dix == NULL || fstat(fileno(_fp_dix), &stats) < 0 ) {
	_error = DNODIX;
	return NULL;
    }

    // Set binary search limits
    low = 0;
    high = (stats.st_size/(2*sizeof(long)))-1;

    // Check against lowest and highest date in index
    if ( GetDixEntry(low*2*sizeof(long), &date, &offset) < 0 ) return NULL;
    if ( date > sp->before ) return urllist;
    if ( GetDixEntry(high*2*sizeof(long), &date, &offset) < 0 ) return NULL;
    if ( date < sp->after ) return urllist;

    // Binary search for lower limit of time interval wanted
    while ( low < high ) {
	split = (low+high)/2;
	if ( GetDixEntry(split*2*sizeof(long), &date, &offset) < 0 )
	    return NULL;

	if ( date >= sp->after ) 
	    high = split;
	else
	    low = split + 1;
    }

    // OK, now we scan the dix for matches and builds URL list
    if ( GetDixEntry(split*2*sizeof(long), &date, &offset) < 0 ) return NULL;

    while ( !feof(_fp_dix) && date <= sp->before ) {

	if ( date >= sp->after ) {
	    unsigned char urlname[MAX_URLLENGTH];
	    unsigned char title[MAX_TITLELENGTH];

	    // Get the URL record and put into list
	    if ( GetURLRecord(offset, &date, urlname, title) < 0 ) {
		_error = DBAD;
		return NULL;
	    }

	    url = new Url(urlname, title, date);
	    urllist->add(url);

	    // Check max list length and abort if too long
	    if ( urllist->length() > sp->max_internal ) {
	    	cout.form(InternalExceeded, sp->max_internal);
	    	exit(1);
	    }
	}

	// Get next dix record
	if ( GetNextDixEntry(&date, &offset) < 0 )
	    if ( error() != DEOF ) return NULL;
    }

    return urllist;
}
    

//
// Write the URL list to URL file
//
int Disk::StoreUrl(URLPTRSplaySet u)
{
    if (_mode == DREAD) {
	_error = DMODE;
	return -1;
    }
	
    long offset = 0;
    int urllen;
    int titlelen;
    
    for (Pix i = u.first(); i != 0 ; u.next(i)) {
	long date;
	
	// Store the offset this information is stored in the url-file in
	// the url object.
	u(i)->setOffset(offset);

	// Write the date to the url-file
	date = u(i)->date();
	fwrite((char *) &date, sizeof(long), 1, _fp_url);
	offset += sizeof(long);

 	// Write the url to the url-file.
	urllen = sizeof(char)*strlen((char *)u(i)->url());
	if ( urllen >= MAX_URLLENGTH ) urllen = MAX_URLLENGTH-1;
	fwrite((char *) u(i)->url(), urllen, 1, _fp_url);
	fwrite("\n", sizeof(char), 1, _fp_url);
	offset += urllen + sizeof(char);

	// Write out the title for the url to the url-file.
	titlelen = sizeof(char)*strlen((char *)u(i)->title());
	if ( titlelen >= MAX_TITLELENGTH ) titlelen = MAX_TITLELENGTH-1;
	fwrite((char *) u(i)->title(), titlelen, 1, _fp_url);
	fwrite("\n", sizeof(char), 1, _fp_url);
	offset += titlelen + sizeof(char);
    }

    return 0;
}


//
// Get next entry, complete with URL data
// Use faked URL data if fake_url set (save disk accesses)
//
Entry* Disk::nextEntry(int initialize)
{
    unsigned char *word;
    unsigned long offset;
    long num;
    unsigned char urlname[MAX_URLLENGTH];
    unsigned char title[MAX_TITLELENGTH];
    unsigned long date;

    // If not in read mode we have a serious problem..
    if (_mode == DWRITE) {
	_error = DMODE;
	return NULL;
    }

    // Initialize as required
    if ( initialize ) {
	if ( GetIndexBlock(0) < 0 ) return NULL;
    }

    // Get the index word and pointer offset
    if ( GetNextIndexEntry(&word, &offset) < 0 ) return NULL;

    Entry *entry = new Entry(word);

    if ( GetUrls(&entry->_urllist, offset, NULL) < 0 ) {
	free(entry);
	return NULL;
    }

    return entry;
}

//
// Get next word and number of occurrences
//
int Disk::nextFrequency(unsigned char* & word, int initialize)
{
    unsigned long offset;
    long num = -1;
    char urlname[MAX_URLLENGTH];
    char title[MAX_TITLELENGTH];

    // Read mode only function
    if (_mode == DWRITE) {
	_error = DMODE;
	return -1;
    }

    // Initialize as required
    if ( initialize ) {
	if ( GetIndexBlock(0) < 0 ) return -1;
    }

    // Get the index word and pointer offset
    if ( GetNextIndexEntry(&word, &offset) < 0 ) return -1;

    // Pull in ptr list length from ptr file
    if ( pointer_fp!=offset && fseek(_fp_ptr, offset, 0) < 0 ) {
	_error = DBAD;
	return -1;
    }

    if ( fread(&num, sizeof(long), 1, _fp_ptr) == 0 ) {
	_error = DBAD;
	return -1;
    }
    pointer_fp = offset + sizeof(long);
	
    return num;
}

//
// Store one entry to disk
//
int Disk::StoreEntry(Entry& e)
{
    int header_space = 0;
    
    if (_mode == DREAD) {
	_error = DMODE;
	return -1;
    }

    // The location of this entry in pointer file
    long offset = pointer_fp;	

    // Set entry
    if ( SetNextIndexEntry(e.word(), offset) < 0 ) return -1;

    // write the numbers of pointers first in the ptr file.
    offset = e._urllist.length();
    fwrite((char *) &offset, sizeof(long), 1, _fp_ptr);
    pointer_fp += sizeof(long);

    // Write the URL pointers to file
    for (Pix n = e._urllist.first(); n != 0 ; e._urllist.next(n)) {
	offset = e._urllist(n)->offset();
	fwrite((char *) &offset, sizeof(long), 1, _fp_ptr);
	pointer_fp += sizeof(long);
    }
    return 0;
}

//
// Flush current index buffer block content to disk
//
int Disk::FlushBuffer()
{
    if (_mode == DREAD) {
	_error = DMODE;
	return -1;
    }

    // If no data we write no block
    if ( _last==0 && _header.blocks==0 ) return 0;

    // write current buffer to disk
    if ( _block == 0 ) 
	memcpy(&_buffer[sizeof(_header)], &_last, sizeof(int));
    else
	memcpy(&_buffer[0], &_last, sizeof(int));
    
    if ( fwrite(_buffer, _header.bsize, 1, _fp_ind) == 0 ) {
	_error = DIO;
	return -1;
    }

    _header.blocks++;
    _block++;
    
    _ind = sizeof(long);
    _last = 0;

    return 0;
}

//
// Get the index header from the start of the header file
//
int Disk::ReadHeader()
{
    // Can't use GetIndexBlock() here.. _buffer is not allocated
    fseek(_fp_ind, 0, 0);
    if (fread(&_header, sizeof(_header), 1, _fp_ind) < 1) {
        _error = DBAD;
        return -1;
    }

    // Make sure the index file is in a well-defined state
    fseek(_fp_ind, 0, 0);
    _block = -1;
    
    // Some primitive checks for header consistency
    // Assumes blocksizes above 64k is bad
    if ( _header.bsize <= 0 || _header.bsize > 64*1024 ) {
	_error = DBAD;
	return -1;
    }
    
    return 0;
}

//
// Write index header to file
//
int Disk::StoreHeader()
{
    fseek(_fp_ind, 0, 0);
    
    if (fwrite(&_header, sizeof(_header), 1, _fp_ind) < 1) {
	_error = DWRITE;
	return -1;
    }

    // Make sure we know where index file pointer is
    fseek(_fp_ind, 0, 0);
    _block = 0;
    
    return 0;
}

//
// Get a given index block from disk
//
int Disk::GetIndexBlock(long blockno)
{
    // We better be in read mode..
    if (_mode != DREAD) {
	_error = DMODE;
	return -1;
    }

    // If block already loaded just set the pointers and return
    if ( _block == blockno ) {
	if ( _block == 0 ) {
		memcpy((char *) &_last, &_buffer[sizeof(_header)], sizeof(int));
		_ind = sizeof(_header)+sizeof(int);
    	} else {
		memcpy((char *) &_last, &_buffer[0], sizeof(int));
		_ind = sizeof(int);
    	}

	return 0;
    }
    
    // Set file pointer and load block
    if ( fseek(_fp_ind, blockno*_header.bsize, 0) < 0 ) {
	_error = DBAD;
	return -1;
    }
    
    _block = blockno - 1;

    return GetNextIndexBlock();
}
    

//
// Read next index block from disk
//
int Disk::GetNextIndexBlock()
{
    // We better be in read mode..
    if (_mode != DREAD) {
	_error = DMODE;
	return -1;
    }

    // Pull in the block
    if ( fread(_buffer, 1, _header.bsize, _fp_ind) == 0  ) {
	_error = DEOF;	
	return -1;
    }

    // Update data structures
    _block++;
    if ( _block == 0 ) {
	memcpy((char *) &_last, &_buffer[sizeof(_header)], sizeof(int));
	_ind = sizeof(_header)+sizeof(int);
    } else {
	memcpy((char *) &_last, &_buffer[0], sizeof(int));
	_ind = sizeof(int);
    }

    return 0;
}

//
// Get next index entry
//
int Disk::GetNextIndexEntry(unsigned char **word, unsigned long *offset)
{
    // Read mode is mandatory..
    if (_mode != DREAD) {
	_error = DMODE;
	return -1;
    }

    // Load new block if we are at end
    if ( _ind>_last && GetNextIndexBlock()<0 ) return -1;
   
    
    // Get the index word
    *word = &_buffer[_ind];
    _ind += strlen((char *) *word);
    _ind++;

    // Get offset into pointer file
    memcpy((char *)offset, &_buffer[_ind], sizeof(long));
    _ind += sizeof(long);

    return 0;
}

//
// Append an index entry to index file
//
int Disk::SetNextIndexEntry(unsigned char *word, unsigned long offset)
{
    // The length of this entry
    int wordsize = strlen((char *)word) + 1;

    // Verify total length is not too big. Max size is limited by block size
    // Entry is silently ignored if not caught by caller.
    if ( wordsize + 2*sizeof(long) >= _header.bsize ) {
	_error = DBIG;
	return -1;
    }
    
    // Check if room in current block, flush block if full
    if ( _ind + wordsize + sizeof(long) >= _header.bsize ) 
	FlushBuffer();

    // Put entry into block
    _last = _ind;
    memcpy(&_buffer[_ind], (char *)word, wordsize);
    _ind += wordsize;
    memcpy(&_buffer[_ind], &offset, sizeof(long));
    _ind += sizeof(long);

    // The index letter for the hash table
    int indl = word[0];
    if (_header.letter[indl] == -1) {
	_header.letter[indl] = _header.blocks;
    }

    return 0;
}
    

//
// Load URLS given offset into pointer file
//
int Disk::GetUrls(URLPTRSplaySet *urllist, long ptr_offset, 
		  struct search_params *sp)
{
    int num;
    int offset;
    unsigned char urlname[MAX_URLLENGTH];
    unsigned char title[MAX_TITLELENGTH];
    long date;   


    // Get pointer list length
    if ( (num = GetNumberOfPointers(ptr_offset)) < 0 ) return -1;

    // Get offsets and corresponding data from URL file
    while (num != 0) {
	fread(&offset, sizeof(long), 1, _fp_ptr);  // fileio system will buffer
	pointer_fp += sizeof(long);
	
	// Get URL data from URL file
	if ( GetURLRecord(offset, &date, urlname, title) < 0 ) return -1;

	if ( sp == NULL || ( sp->after <= date && sp->before >= date )) {
	    Url *url = new Url(urlname, title, date);
	    url->setOffset(offset);
	    urllist->add(url);
	}
	
	num--;
    }

    return 0;
}

//
// Get number of pointers for an entry, i.e. list length
// Will also set file pointer
//
int Disk::GetNumberOfPointers(unsigned long offset)
{
    int num;
    
    // Get pointer list length
    if ( offset != pointer_fp && fseek(_fp_ptr, offset, 0) < 0 ) {
	_error = DBAD;
	return -1;
    }
    
    if ( fread(&num, sizeof(long), 1, _fp_ptr) == 0 ) {
	_error = DBAD;
	return -1;
    }

    pointer_fp = offset + sizeof(long);
    
    return num;
}

//
// Read one URL record given offset
//
int Disk::GetURLRecord(unsigned offset, long *date, unsigned char *urlname,
		 unsigned char *title)
{
    if ( fseek(_fp_url, offset, 0) < 0 ) {
	_error = DBAD;
	return -1;
    }

    return GetNextURLRecord(date, urlname, title);
}

//
// Read one record from the URL database file
// Returns record length
//
int Disk::GetNextURLRecord(long *date, unsigned char *urlname,
			unsigned char *title)
{
    int titlelen, urllen;
    int totallength = 0;
    
    if ( fread((char *) date, sizeof(long), 1, _fp_url) == 0) {
	_error = DEOF;
	return -1;
    }

    totallength += sizeof(long);
    
    // Get the URL
    if ( fgets((char *) urlname, MAX_URLLENGTH, _fp_url) == 0) {
	_error = DBAD;
	return -1;
    }
    urllen = strlen((char *)urlname);
    totallength += urllen;
	
    // If url was too long read off extras
    if ( urllen == MAX_URLLENGTH-1 && urlname[MAX_URLLENGTH-2] != '\n' ) 
	do { totallength++; } while ( getc(_fp_url) != '\n' );
    
    
    if ( fgets((char *) title, MAX_TITLELENGTH, _fp_url) == 0) {
	_error = DBAD;
	return -1;
    }
    titlelen = strlen((char *)title);
    totallength += titlelen;

    // If title was too long read off extras
    if ( titlelen == MAX_TITLELENGTH-1 && title[MAX_TITLELENGTH-2] != '\n' ) 
	do { totallength++; } while ( getc(_fp_url) != '\n' );
	
    urlname[urllen-1] = '\0';
    title[titlelen-1] = '\0';

    return totallength;
}

//
// Construct a new pointer list from two lists
// This function is used specifically by Merge()
// pointer_fp's are updated by Merge()
// On entry it is assumed the file1 and file2 file pointers
// points to the beginning of the lists
//
int Disk::CopyPointers(int num1, FILE *file1, int num2, FILE *file2,
		 unsigned long offset2)
{
#define PTRBUFSIZE 2048
    unsigned long ptrbuf[PTRBUFSIZE];
    long total_length = num1 + num2;

    // Write new total length
    if ( fwrite(&total_length, sizeof(long), 1, _fp_ptr) < 0 ) {
	_error = DIO;
	return -1;
    }

    // Copy file1 data
    while ( num1 > 0 ) {
	int chunk = (num1>PTRBUFSIZE) ? PTRBUFSIZE : num1;
	if ( fread(ptrbuf, sizeof(long), chunk, file1) < 0 ) {
	    _error = DIO;
	    return -1;
	}

	if ( fwrite(ptrbuf, sizeof(long), chunk, _fp_ptr) < 0 ) {
	    _error = DIO;
	    return -1;
	}

	num1 -=chunk;
    }

    // Copy file2 data
    while ( num2 > 0 ) {
	int chunk = (num2>PTRBUFSIZE) ? PTRBUFSIZE : num2;
	if ( fread(ptrbuf, sizeof(long), chunk, file2) < 0 ) {
	    _error = DIO;
	    return -1;
	}

	// Add offsets for 2nd file
	for ( int i=0; i<chunk; i++) ptrbuf[i] += offset2;
	
	if ( fwrite(ptrbuf, sizeof(long), chunk, _fp_ptr) < 0 ) {
	    _error = DIO;
	    return -1;
	}

	num2 -=chunk;
    }

    return 0;
}
   
//
// Build the date index by scanning the URL tree
// URL file must be saved so offsets are present
//
int Disk::MakeDateIndex(Index &index)
{
    unsigned long offset;
    URLPTRSplaySet dateindex;

    // Write mode is mandatory..
    if (_mode != DWRITE) {
	_error = DMODE;
	return -1;
    }

    // Build the date index tree from the URL tree
    unsigned char fakeurl[25];
    Pix i;
    for (i = index._url.first(); i != 0 ; index._url.next(i)) {
	sprintf((char *)fakeurl, "%012d%012d", index._url(i)->date(),
		index._url(i)->offset());

	Url *url = new Url(fakeurl, (unsigned char *)"",
			   index._url(i)->date());
	url->setOffset(index._url(i)->offset());
	dateindex.add(url);
    }

    // Dump the index tree to file
    for (i = dateindex.first(); i != 0 ; dateindex.next(i)) {
	if ( SetNextDixEntry(dateindex(i)->date(), dateindex(i)->offset()) < 0) {
	    return -1;
	}
    }

    fclose(_fp_dix);

    return 0;
}
    
//
// Read date index entry given offset
//
int Disk::GetDixEntry(unsigned long dix_offset, long *date,
		      unsigned long *url_offset)
{
    if ( fseek(_fp_dix, dix_offset, 0) < 0 ) {
	_error = DBAD;
	return -1;
    }

    return GetNextDixEntry(date, url_offset);
}

//
// Read next date index 
//
int Disk::GetNextDixEntry(long *date, unsigned long *url_offset)
{
    if ( fread(date, sizeof(long), 1, _fp_dix) <= 0 ) {
	_error = DEOF;
	return -1;
    }

    if ( fread(url_offset, sizeof(long), 1, _fp_dix) <= 0 ) {
	_error = DEOF;
	return -1;
    }

    return 0;
}

//
// Write a dix entry
//
int Disk::SetNextDixEntry(long date, unsigned long url_offset)
{
    if ( fwrite(&date, sizeof(long), 1, _fp_dix) <= 0 ) {
	_error = DIO;
	return -1;
    }
    if ( fwrite(&url_offset, sizeof(unsigned long), 1, _fp_dix) <= 0 ) {
	_error = DIO;
	return -1;
    }

    return 0;
}

//
// Constructor
//
Disk::Disk(unsigned char *s, int mode)
{
    unsigned char *filename;

    _mode = mode;
    _error = 0;
    _header.bsize = BLOCK_SIZE;
    _header.blocks = 0;

    int len = strlen((char *) s);
    _indexname = (unsigned char *) malloc(sizeof(char)*(len+1));
    filename = (unsigned char *) malloc(sizeof(char)*(len+5));
    
    strcpy((char *) _indexname, (char *) s);
       
    if (mode == DWRITE) {
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".url", 4);

	if ((_fp_url = fopen((char *) filename, "w")) == NULL) {
	    _error = DOPEN;
	    return;
	}

	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".ind", 4);

	if ((_fp_ind = fopen((char *) filename, "w")) == NULL) {
	    _error = DOPEN;
	    return;
	}

	// Open pointers file
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".ptr", 4);

	if ((_fp_ptr = fopen((char *) filename, "w")) == NULL) {
	    _error = DOPEN;
	    return;
	}

	// Open date index file
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".dix", 4);

	if ((_fp_dix = fopen((char *) filename, "w")) == NULL) {
	    _error = DOPEN;
	    return;
	}


	// Prepare data structures
	_block = 0;	   	// After open first block is ready..
	_ind = sizeof(int) + sizeof(_header);
	_last = 0;
	pointer_fp = 0;
	
	for (int i = 0; i < 256 ; i++) {
	    _header.letter[i] = -1;
	}
    } 

    if (mode == DREAD) {
	// Open url file
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".url", 4);
	
	if ((_fp_url = fopen((char *) filename, "r")) == NULL) {
	    _error = DOPEN;
	    return;
	} 

	// Open main index file
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".ind", 4);	

	if ((_fp_ind = fopen((char *) filename, "r")) == NULL) {
	    _error = DOPEN;
	    return;
	}

	// Open pointers file
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".ptr", 4);

	if ((_fp_ptr = fopen((char *) filename, "r")) == NULL) {
	    _error = DOPEN;
	    return;
	}

	// Open date index file.. not mandatory..
	strcpy((char *) filename, (char *) _indexname);
	strncat((char *) filename, ".dix", 4);

	if ((_fp_dix = fopen((char *) filename, "r")) == NULL) {
	    _fp_dix = NULL;
	}

	// Prepare data structures
	if (ReadHeader() == -1) return;
	_block = -1;		// -1 == no block in buffer
	_ind = 1;		// _ind > _last -> load new block
	_last = 0;
	pointer_fp = 0;
    }

    _buffer = new unsigned char[_header.bsize];
}

Disk::~Disk()
{
    delete _buffer;
    fclose(_fp_ind);
    fclose(_fp_ptr);
    fclose(_fp_url);
}

